;;;;
;;;; c e e v a l . s c m        -- A very simple and stupid reflective
;;;;                               evaluator for Scheme
;;;;                               in continuation- with environment-
;;;;                               passing style
;;;;
;;;; Copyright C 1997 Fare Rideau under the GPL
;;;;
;;;;           Author: Fare Rideau [rideau@unice.fr]

;;;; See README for description

(provide 'ceeval)
(require 'macrolib)
(require 'lib)
(require 'tags)
(require 'debug)

(*test-off*)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Scheme objects
;;;;

;;;; using the tags.scm library, let's have user-defined types

(define lambda-tag  (make-tag 'lambda))
(define lambda?     (tagged-object-detector lambda-tag))
(define make-lambda (tagged-object-maker lambda-tag))

(define appl-tag  (make-tag 'appl))
(define appl?     (tagged-object-detector appl-tag))
(define make-appl (tagged-object-maker appl-tag))

(define reflective-tag  (make-tag 'reflective))
(define reflective?     (tagged-object-detector reflective-tag))
(define make-reflective (tagged-object-maker reflective-tag))

;(define set!-tag  (make-tag 'set!))
;(define set!?     (tagged-object-detector set!-tag))
;(define make-set! (tagged-object-maker set!-tag))



;;;; The usual definition for a Scheme constant

(define (constant? x)
  (or (boolean? x)
      (null? x)
      (number? x)
      (char? x)))


(define (atom? x)
  (or (constant? x)
      (id? x)))				; an id is a global memory location


;;;; constructing and deconstructing scheme objects

(define (constructor-of x)
  (cond
   ((atom? x)   (lambda () x))
   ((pair? x)   cons)
   ((vector? x) vector)
   ((lambda? x) make-lambda)
;  ((set!? x)   make-set!)
   ((appl? x)   make-appl)
   (else        (error))))

(define (body-of x)
  (cond
   ((atom? x)   '())
   ((pair? x)   (list (car x) (cdr x)))
   ((vector? x) (vector->list x))
   ((lambda? x) (cdr x))
   ((appl? x)   (cdr x))
;  ((set!? x)   (cdr x))
   (else        (error "object of unknown body" x))))

(define (bindings-of x)
  (cond
   ((lambda? x) (fringe (cadr x)))
   (else        '())))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; an "id" datatype, to represent named environment/memory locations...
;;;;

(define id::tag (make-tag 'id))
(define id::counter 0)
(define (id::count)
  (set! id::counter (+ id::counter 1))
  id::counter)

(define (make-id sym)
  (list id::tag sym (id::count)))
(define (id? x) (or (symbol? x) (and (list? x) (eq? (car x) id::tag))))
(define (id->symbol id)
  (assert id? id)
  (if (symbol? id) id (cadr id)))
(define (id->count id)
  (assert id? id)
  (if (symbol? id) 0 (caddr id)))
(define (alpha-convert-id id)
  (make-id (id->symbol id)))
(define (genid)
  (make-id 'G))

; To print an id
(define (id->string id)
  (assert id? id)
  (let* ((str (symbol->string (id->symbol id)))
	 (len (string-length str))
	 (bad (and (< 0 len)
		   (member (string-ref str (- len 1))
			    '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\_))))
	 (mod (if bad "_" ""))
	 (cnt (id->count id))
	 (suf (if (zero? cnt) "" (number->string cnt))))
    (string-append str mod suf)))

; unique id->symbol mapping
(define (id->name id)
  (string->symbol (id->string id)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Environments as global memory store
;;;;

;;;; simulate abstract data type (see tags.scm)

(define env-tag  (make-tag 'env))
(define env?     (tagged-object-detector env-tag))
(define make-env (tagged-object-maker env-tag))


;;;; Abstract operations on environments

(define (env:bindings env)
  (if (env? env) (cadr env)
    (error "env:binding requires an environment, got " env)))

(define (env:add-binding env b)
  (apply make-env (cons b (env:bindings env)) (cddr env)))

(define (env:find-binding env symbol ok nok)
  (let ((binding (assq symbol (env:bindings env))))
     (if binding (ok binding) (nok))))

(*test*
   ((make-env '((a b) (c d))) (((magic) env) ((a b) (c d))))
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Explicit Substitutions
;;;;

; A substitution is an a-list of (replaced-symbol . replacing-value) 

(define (replace v e x)
  "Replace a variable v by expression e in expression x"
  (if (atom? x) (if (eq? x v) e x)
    (let ((c (constructor-of x))
	  (w (bindings-of x))
	  (b (body-of x))
          (r (lambda (x) (replace v e x))))
      (if (member v w) x
	(apply c (map r b))))))

(define (do-apply-subst subst atom)
  "Apply a substitution on an atom"
  (let ((foo (assq atom subst)))
    (if foo (cdr foo) atom)))

(define (apply-subst subst expr)
  "Apply a substitution on an expression"
  (if (atom? expr) (do-apply-subst subst expr)
    (let* ((c (constructor-of expr))
	   (bindings (bindings-of expr))
	   (body (body-of expr))
	   (off (lambda (w) (not (member (car w) bindings))))
           (rsubst (filter off subst))
           (recurse (lambda (x) (apply-subst rsubst x))))
      (apply c (map recurse body)))))

(*test*
  ((replace 'x '(+ x 1) '(* 3 x x))
   (* 3 (+ x 1) (+ x 1)))
  ((apply-subst '((x . y) (y . x)) '(* x (+ y x)))
   (* y (+ x y)))
)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Alpha conversion
;;;;

; rename the outer variables of a lambda- or quantified expression
; to do before to apply the lambda-expression in a dynamic way.

(define (primitive-alpha-convert x)
  (if (atom? x) x
    (let* ((c (constructor-of x))
	   (b (body-of x))
	   (w (bindings-of x))
	   (v->s (lambda (v) (cons v (alpha-convert-id v))))
	   (subst (map v->s w))
	   (do-subst (lambda (x) (apply-subst subst x))))
      (apply c (map do-subst b)))))

(*test* ((primitive-alpha-convert (make-lambda '(x) 'x)))
	(((magic) lambda) ((((magic) id) x 1)) (((magic) id) x 1)))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Special forms
;;;;

(define orig-quote
  (make-reflective
   (lambda (cont env arg . rest)
     (cont env arg))))

(define orig-lambda
  (make-reflective
   (lambda (cont env . args)
     (cont env (apply make-lambda args)))))

;(define orig-set!
;  (make-reflective
;   (lambda (cont env id val)
;     (cont (env:add-binding env (cons id val)) #f))))

(define orig-lambda-reflect
  (make-reflective
   (lambda (cont env . args)
     (cont env make-reflective (apply make-lambda args)))))

(define orig-macro
  (make-reflective
   (lambda (cont env formals . body)
     (cont env
       (make-reflective
	(lambda (runcont runenv . args)
	  ((lambda (macrocont)
	     (cee:do-apply macrocont runenv
	                   (apply make-lambda formals body) args))
	   (lambda (runenv expr)
	     (ceeval runcont runenv expr)))))))))

;(define orig-current-env
;  (make-reflective
;   (lambda (cont env)
;     (cont env (env:bindings env)))))

;(define orig-call/cc
;  (make-reflective
;   (lambda (cont env fun)
;     ((lambda (cont)
;	(ceeval cont env fun))
;      (lambda (env efun)
;	(cee:do-apply cont env efun (list cont)))))))

(define special-forms-bindings
  `((lambda           . ,orig-lambda)
     (quote           . ,orig-quote)
     (lambda-reflect  . ,orig-lambda-reflect)
     (make-reflective . ,make-reflective)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Reflected Primitives
;;;;

(define-macro (declare-primitive prim)
  `(cons ',prim ,prim))

(define primitives-bindings
  (map-macro (list) declare-primitive
     read write
     eq? memq assq
     equal? member assoc
     eqv? memv assv
     car cdr cons pair?
     null? list append map
     caar caaar caaaar cdaaar
          cdaar cadaar cddaar
     cdar cadar caadar cdadar
          cddar caddar cdddar
     cadr caadr caaadr cdaadr
          cdadr cadadr cddadr
     cddr caddr caaddr cdaddr
          cdddr cadddr cddddr
   ))

;(write primitives-bindings)
;(newline)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Initial settings for the evaluator
;;;;
(define orig-env
  (make-env
   (append '((foo . bar) (bar . quux))
	   special-forms-bindings
	   primitives-bindings)))

(define (orig-cont env expr)
  (display "Computation terminated!\nResult: ")
  (write expr)
  (newline))

(define (orig-eval expr)
  (call-with-current-continuation
   (lambda (cont)
     (ceeval (lambda (env expr) (cont expr)) orig-env expr))))

(define null-env  (make-env '()))
(define null-cont
  (lambda args
    (error "This continuation should never be reached; got args " args)))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Evaluation
;;;;

; And now for the real thing...

(define (ceeval cont env expr)
  (cond
   ((constant? expr) (cont env expr))
   ((id? expr) (cee:symbol cont env expr))
   ((pair? expr) (cee:apply cont env (car expr) (cdr expr)))
   (else (error "syntax error in " expr))))

(define (cee:symbol cont env symbol)
  (env:find-binding env symbol
     (lambda (binding) (cont env (cdr binding)))
     (lambda () (error "unbound variable " symbol))))

(define (cee:apply cont env fun args)
  ((lambda (cont) (ceeval cont env fun))
   (lambda (env efun)
     (if (reflective? efun)
	 (cee:do-apply null-cont null-env (cadr efun) `(,cont ,env ,@args))
       (cee:apply-normal cont env efun args)))))

(define (left-to-right cont env accum args)
  (if (null? args)
      (cont env '())
    ((lambda (cont)
       (ceeval cont env (car args)))
     (lambda (env ecarargs)
       ((lambda (cont)
	  (left-to-right cont env accum (cdr args)))
	(lambda (env ecdrargs)
	  (cont env (accum ecarargs ecdrargs))))))))

(define	(right-to-left cont env accum args)
  (if (null? args)
      (cont env '())
    ((lambda (cont)
       (right-to-left cont env accum (cdr args)))
     (lambda (env ecdrargs)
       ((lambda (cont)
	  (ceeval cont env (car args)))
	(lambda (env ecarargs)
	  (cont env (accum ecarargs ecdrargs))))))))

(define (keep-last a d)
  (if (pair? d) d (list a)))

(define (cee:apply-normal cont env efun args)
  ((lambda (cont)
     (left-to-right cont env cons args))
   (lambda (env eargs)
     (cee:do-apply cont env efun eargs))))

(define (cee:do-apply cont env efun eargs)
  (cond
   ((procedure? efun) (cont env (apply efun eargs)))
   ((lambda? efun) (lambda-evaluator cont env efun eargs))
   (else (error "cannot apply " efun " to " eargs))))

(define (lambda-evaluator cont env efun eargs)
  (let ((afun (primitive-alpha-convert efun)))
    (left-to-right
      (lambda (env results) (cont env (car results)))
      (bind-formals env (cadr afun) eargs)
      keep-last
      (cddr afun))))


(define (bind-formals env formals args)
; redundant parameter names not detected!
  (cond
   ((null? formals) (if (null? args) env
		      (error "no argument expected, got " args)))
   ((id? formals)   (env:add-binding env (cons formals args)))
   ((and (pair? formals) (id? (car formals)))
       (if (pair? args)
	   (env:add-binding
	      (bind-formals env (cdr formals) (cdr args))
	      (cons (car formals) (car args)))
	 (error "not enough arguments; expected " formals " got " args)))
   (else (error "wrong format for formals " formals))))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Let's test the evaluator on some easy cases...
;;;;

(*test-on*)
(*test*
  ((orig-eval 1) 1)
  ((orig-eval 'foo) bar)
  ((orig-eval 'bar) quux)
  ((orig-eval '(car (cons 1 2))) 1)
  ((orig-eval '(lambda (foo) foo)) (((magic) lambda) (foo) foo))
  ((orig-eval '((lambda (foo) foo) bar)) quux)
  ((orig-eval '((lambda (foo) (foo bar)) (lambda (foo) foo))) quux)
  ((orig-eval ''(foo . bar)) (foo . bar))
  ((orig-eval '((lambda foo foo) 'foo foo)) (foo bar))
  ((orig-eval '((lambda (foo) (foo))
		((lambda (foo) (lambda () foo))
		 foo))) bar)
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; To finish, a read-eval-print loop
;;;;

((lambda (repl env) (repl repl env))
 (lambda (repl env)
  ((lambda (cont form)
     (if (not (eof-object? form))
	 (ceeval cont env form)))
   (lambda (env val)
     (write val)
     (newline)
     (repl repl env))
   (read)))
 orig-env)
