;;;;
;;;; c e e v a l . s c m        -- A very simple and stupid reflective
;;;;                               evaluator for Scheme
;;;;                               in continuation- with environment-
;;;;                               passing style
;;;;
;;;; Copyright C 1997 Fare Rideau under the GPL
;;;;
;;;;           Author: Fare Rideau [rideau@unice.fr]

;;;; See README for description

(provide 'ceeval)
(require 'macrolib)
(require 'lib)
(require 'tags)
(require 'debug)

(*test-off*)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Scheme objects
;;;;

;;;; using the tags.scm library, let's have user-defined types

(define lambda-tag  (make-tag 'lambda))
(define lambda?     (tagged-object-detector lambda-tag))
(define make-lambda (tagged-object-maker lambda-tag))

;(define appl-tag  (make-tag 'appl))
;(define appl?     (tagged-object-detector appl-tag))
;(define make-appl (tagged-object-maker appl-tag))

(define reflective-tag  (make-tag 'reflective))
(define reflective?     (tagged-object-detector reflective-tag))
(define make-reflective (tagged-object-maker reflective-tag))

(define set!-continue-tag  (make-tag 'set!-continue))
(define set!c?             (tagged-object-detector set!-continue-tag))
(define make-set!c         (tagged-object-maker set!-continue-tag))
; ((make-set!c cont id body) val) binds id to val
; then continues evaluating the body with cont

(define apply-continue-tag  (make-tag 'apply-continue))
(define applyc?             (tagged-object-detector apply-continue-tag))
(define make-applyc         (tagged-object-maker apply-continue-tag))
; ((make-applyc cont args) val)
; applies the values to the arguments and continues with cont

(define eapply-continue-tag  (make-tag 'eapply-continue))
(define eapplyc?             (tagged-object-detector eapply-continue-tag))
(define make-eapplyc         (tagged-object-maker eapply-continue-tag))
;; ((make-eapplyc cont efun eargs args) val)








;;;; The usual definition for a Scheme constant

(define (constant? x)
  (or (boolean? x)
      (null? x)
      (number? x)
      (string? x)		; maybe move out of constant?
      (char? x)))
      
(define (self-eval? x)
  (or (constant? x)
      (procedure? x)
      (reflective? x)
      (lambda? x)
      (set!c? x)
      (applyc? x)
      (eapplyc? x)))

(define (atom? x)
  (or (constant? x)
      (procedure? x)            ; internal procedures are atomic objects...
      (id? x)))			; an id is a global memory location


;;;; constructing and deconstructing scheme objects
; useful for substitution, GC, etc.

(define (constructor-of x)
  (cond
   ((atom? x)       (lambda () x))
   ((pair? x)       cons)
   ((vector? x)     vector)
   ((lambda? x)     make-lambda)
;  ((appl? x)       make-appl)
   ((set!c? x)      make-set!c)
   ((applyc? x)     make-applyc)
   ((eapplyc? x)    make-eapplyc)
   ((reflective? x) make-reflective)
   (else            (error "object of unknown constructor" x))))

(define (body-of x)
  (cond
   ((atom? x)       '())
   ((pair? x)       (list (car x) (cdr x)))
   ((vector? x)     (vector->list x))
   ((lambda? x)     (cdr x))
;  ((appl? x)       (cdr x))
   ((set!c? x)      (cdr x))
   ((applyc? x)     (cdr x))
   ((eapplyc? x)    (cdr x))
   ((reflective? x) (cdr x))
   (else            (error "object of unknown body" x))))

(define (bindings-of x)
  (cond
   ((lambda? x) (fringe (cadr x)))
   (else        '())))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; an "id" datatype, to represent named environment/memory locations...
;;;;

(define id::tag (make-tag 'id))
(define id::counter 0)
(define (id::count)
  (set! id::counter (+ id::counter 1))
  id::counter)

(define (make-id sym)
  (list id::tag sym (id::count)))
(define (id? x)
  (or (symbol? x)
      ((tagged-object-detector id::tag) x)))
(define (id->symbol id)
  (assert id? id)
  (if (symbol? id) id (cadr id)))
(define (id->count id)
  (assert id? id)
  (if (symbol? id) 0 (caddr id)))
(define (alpha-convert-id id)
  (make-id (id->symbol id)))
(define (genid)
  (make-id 'G))

; To print an id
(define (id->string id)
  (assert id? id)
  (let* ((str (symbol->string (id->symbol id)))
	 (len (string-length str))
	 (bad (and (< 0 len)
		   (member (string-ref str (- len 1))
			    '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\_))))
	 (mod (if bad "_" ""))
	 (cnt (id->count id))
	 (suf (if (zero? cnt) "" (number->string cnt))))
    (string-append str mod suf)))

; unique id->symbol mapping
(define (id->name id)
  (string->symbol (id->string id)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Environments as global memory store
;;;;

;;;; simulate abstract data type (see tags.scm)

(define env-tag  (make-tag 'env))
(define env?     (tagged-object-detector env-tag))
(define make-env (tagged-object-maker env-tag))


;;;; Abstract operations on environments

(define (env:bindings env)
  (if (env? env) (cadr env)
    (error "env:binding requires an environment, got " env)))

(define (env:add-binding env b)
  (apply make-env (cons b (env:bindings env)) (cddr env)))

(define (env:find-binding env symbol ok nok)
  (let ((binding (assq symbol (env:bindings env))))
     (if binding (ok binding) (nok))))

(*test*
   ((make-env '((a b) (c d))) (((magic) env) ((a b) (c d))))
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Explicit Substitutions
;;;;

; A substitution is an a-list of (replaced-symbol . replacing-value) 

(define (replace v e x)
  "Replace a variable v by expression e in expression x"
  (if (atom? x) (if (eq? x v) e x)
    (let ((c (constructor-of x))
	  (w (bindings-of x))
	  (b (body-of x))
          (r (lambda (x) (replace v e x))))
      (if (member v w) x
	(apply c (map r b))))))

(define (do-apply-subst subst atom)
  "Apply a substitution on an atom"
  (let ((foo (assq atom subst)))
    (if foo (cdr foo) atom)))

(define (apply-subst subst expr)
  "Apply a substitution on an expression"
  (if (atom? expr) (do-apply-subst subst expr)
    (let* ((c (constructor-of expr))
	   (bindings (bindings-of expr))
	   (body (body-of expr))
	   (off (lambda (w) (not (member (car w) bindings))))
           (rsubst (filter off subst))
           (recurse (lambda (x) (apply-subst rsubst x))))
      (apply c (map recurse body)))))

(*test*
  ((replace 'x '(+ x 1) '(* 3 x x))
   (* 3 (+ x 1) (+ x 1)))
  ((apply-subst '((x . y) (y . x)) '(* x (+ y x)))
   (* y (+ x y)))
)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Alpha conversion
;;;;

; rename the outer variables of a lambda- or quantified expression
; to do before to apply the lambda-expression in a dynamic way.

(define (primitive-alpha-convert x)
  (if (atom? x) x
    (let* ((c (constructor-of x))
	   (b (body-of x))
	   (w (bindings-of x))
	   (v->s (lambda (v) (cons v (alpha-convert-id v))))
	   (subst (map v->s w))
	   (do-subst (lambda (x) (apply-subst subst x))))
      (apply c (map do-subst b)))))

(*test* ((primitive-alpha-convert (make-lambda '(x) 'x)))
	(((magic) lambda) ((((magic) id) x 1)) (((magic) id) x 1)))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Special forms
;;;;

(define orig-quote
  (make-reflective
   (lambda (cont env . args)
     (apply cee:continue cont env args))))

(define (make-quote foo)
  (list orig-quote foo))

(define orig-lambda
  (make-reflective
   (lambda (cont env . args)
     (cee:continue cont env (apply make-lambda args)))))

(define (args-for-apply . args)
   (letrec ((accumulate-args
	      (lambda (args rest)
		 (if (null? (cdr rest)) (append args (car rest))
		     (accumulate-args (append args (list (car rest))) (cdr rest))))))
	   (accumulate-args '() args)))

(define orig-apply
   (make-reflective
     (lambda (cont env fun . args)
        (cee:continue
	    (make-eapplyc
	     (lambda (env expr)
	       (cee:eeapply (car expr) env (cadr expr) (cddr expr)))
	     args-for-apply
	     '() (cons fun args))
	    env cont))))


(define orig-set!
  (make-reflective
   (lambda (cont env id expr)
     (ceeval (make-set!c cont id #f) env expr))))

(define orig-lambda-reflect
  (make-reflective
   (lambda (cont env . args)
     (cee:continue cont env (make-reflective (apply make-lambda args))))))

(define orig-current-env
  (make-reflective
   (lambda (cont env)
     (cee:continue cont env (env:bindings env)))))

(define (do-macro cont env lam . args)
  ; This is an ugly hack to never absorb continuations,
  ; so an explicit GC could work...
  (cee:eeapply
;   (lambda (env expr) (display "Macro expanded to: ") (write expr) (newline);;;;;
;    (cee:continue;;;;;
   (make-eapplyc
	    (lambda (env expr)
	       (ceeval (car expr) env (cdr expr)))
	    cons (list cont) '())
;     env expr));;;;
   env lam args))

(define orig-macro
  (make-reflective
   (lambda (cont env formals . body)
     (cee:continue cont env
       (make-reflective	do-macro (apply make-lambda formals body))))))

(define orig-call/cc
  (make-reflective
   (lambda (cont env fun)
     (ceeval (make-applyc cont (make-quote cont)) env fun))))

(define special-forms-bindings
  `((lambda           . ,orig-lambda)
     (quote           . ,orig-quote)
     (set!            . ,orig-set!)
     (lambda-reflect  . ,orig-lambda-reflect)
     (extensional-if  . ,(lambda (b t f) (if b t f)))
     (call/cc         . ,orig-call/cc)
     (macro           . ,orig-macro)
     (apply           . ,orig-apply)
     (make-reflective . ,make-reflective)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Reflected Primitives
;;;;

(define-macro (declare-primitive prim)
  `(cons ',prim ,prim))

(define primitives-bindings
  (map-macro (list) declare-primitive
     read write display newline
     eq? memq assq
     equal? member assoc
     eqv? memv assv
     car cdr cons pair?
     null? list append
     caar caaar caaaar cdaaar
          cdaar cadaar cddaar
     cdar cadar caadar cdadar
          cddar caddar cdddar
     cadr caadr caaadr cdaadr
          cdadr cadadr cddadr
     cddr caddr caaddr cdaddr
          cdddr cadddr cddddr
     + - * / = < <= > >=
     exit
   ))

;(write primitives-bindings)
;(newline)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Initial settings for the evaluator
;;;;
(define orig-env
  (make-env
   (append '((foo . bar) (bar . quux))
	   special-forms-bindings
	   primitives-bindings)))

(define (orig-cont env expr)
;  (display "Computation terminated!\nResult: ")
;  (write expr)
;  (newline)
  expr)

(define (orig-eval expr)
  (ceeval orig-cont orig-env expr))

(define null-env  (make-env '()))
(define null-cont
  (lambda args
    (error "This continuation should never be reached; got args " args)))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Evaluation
;;;;

; And now for the real thing...

(define (ceeval cont env expr)
  (cond
   ((self-eval? expr) (cee:continue cont env expr))
   ((id? expr)       (cee:symbol cont env expr))
   ((pair? expr)     (cee:apply cont env (car expr) (cdr expr)))
   (else             (error "syntax error in " expr))))

(define (cee:symbol cont env symbol)
  (env:find-binding env symbol
     (lambda (binding) (cee:continue cont env (cdr binding)))
     (lambda () (error "unbound variable " symbol))))

(define (cee:continue cont env . values)
  (cond
   ((procedure? cont) (apply cont env values))
   ((set!c? cont)
      (ceeval (cadr cont)
                    (if (caddr cont)
		    (env:add-binding env (cons (caddr cont) (car values))))
		    (cadddr cont)))
    ((applyc? cont)
      (cee:eapply (cadr cont) env (car values) (caddr cont)))
    ((eapplyc? cont)
      (cee:do-eapply (cadr cont) env (caddr cont)
                     (append (cadddr cont) values) (car (cddddr cont))))
    ((lambda? cont)
      (cee:eeapply null-cont env cont values))
   (else (error "bad continuation" cont))))

(define (cee:apply cont env fun args)
  (ceeval (make-applyc cont args) env fun))

(define (cee:eapply cont env efun args)
  (if (reflective? efun)
      ((lambda (rfun eargs)
	(if (procedure? rfun) (apply rfun eargs)
            (eeapply cont env rfun eargs)))
       (cadr efun) (append (list cont env) (cddr efun) args))
      (cee:do-eapply cont env efun '() args)))

(define (cee:do-eapply cont env efun eargs args)
   (if (null? args)
       (cee:eeapply cont env efun eargs)
       (ceeval (make-eapplyc cont efun eargs (cdr args))
               env (car args))))

(define (cee:eeapply cont env efun eargs)
  (cond
   ((lambda? efun)     (cee:lambda-evaluator cont env efun eargs))
   ((procedure? efun)
;     (write (list "cee:eeapply" efun eargs))(newline)
     (cee:continue cont env (apply efun eargs)))
   ((set!c? efun)      (apply cee:continue efun env eargs))
   ((applyc? efun)     (apply cee:continue efun env eargs))
   ((eapplyc? efun)    (apply cee:continue efun env eargs))
   ((reflective? efun) (cee:eapply cont env efun
			 (map (lambda (x) (list orig-quote x)) eargs)))
   (else (error "cannot apply" efun "to" eargs))))

(define (cee:lambda-evaluator cont env efun eargs)
  (let* ((afun      (primitive-alpha-convert efun))
	 (env       (bind-formals env (cadr afun) eargs))
	 (bodyfirst (caddr afun))
         (bodyrest  (cdddr afun)))
    (if (null? bodyrest)
	(ceeval cont env bodyfirst)
      (cee:do-eapply cont env
		     (apply make-lambda (genid) bodyrest) '()
		     (list bodyfirst)))))

(define (bind-formals env formals eargs)
; redundant parameter names not detected!
  (cond
   ((null? formals) (if (null? eargs) env
		      (error "no argument expected, got " eargs)))
   ((id? formals)   (env:add-binding env (cons formals eargs)))
   ((and (pair? formals) (id? (car formals)))
       (if (pair? eargs)
          (bind-formals
 	    (env:add-binding env (cons (car formals) (car eargs)))
	    (cdr formals) (cdr eargs))
	 (error "not enough arguments; expected" formals "got" eargs)))
   (else (error "wrong format for formals " formals))))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Let's test the evaluator on some easy cases...
;;;;

;(*test-on*)
(*test*
  ((orig-eval 1) 1)
  ((orig-eval 'foo) bar)
  ((orig-eval 'bar) quux)
  ((orig-eval '(car (cons 1 2))) 1)
  ((orig-eval '(lambda (foo) foo)) (((magic) lambda) (foo) foo))
  ((orig-eval '((lambda (foo) foo) bar)) quux)
  ((orig-eval '((lambda (foo) (foo bar)) (lambda (foo) foo))) quux)
  ((orig-eval ''(foo . bar)) (foo . bar))
  ((orig-eval '((lambda foo foo) 'foo foo)) (foo bar))
  ((orig-eval '((lambda (foo) (foo))
		((lambda (foo) (lambda () foo))
		 foo))) bar)
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; To finish, a read-eval-print loop
;;;;

((lambda (repl env) (repl repl env))
 (lambda (repl env)
  ((lambda (cont form)
     (if (eof-object? form)
         (map display `("Used " ,(length (env:bindings env)) " memory cells\n"))
	 (ceeval cont env form)))
   (lambda (env val)
     (display ">> ")
     (write val)
     (newline)
     (repl repl env))
   ((lambda () (display "")(read)))))
 orig-env)
