;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;; Scheme support code for Tunes ;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; This is a standalone library for
; simple scheme structures,
; thoroughly used to bootstrap the HLL interpreter...

(provide 'lib)
(require 'debug)

;;;;;;;;;;;;;;;;;;;;;;;;;; Higher-Order functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;; Trivial functions...
(define (Id x) x)
(define (->const x) (lambda dummy x))
(define nop (->const #f))

;;;;; Folding combinators

; Fold right. Ex: (foldr 0 + '(3 2 6)) = (+ 3 (+ 2 (+ 6 0)))
(define (foldr zero fun l)
  (if (null? l)
      zero
      (fun (car l) (foldr zero fun (cdr l)))))

; Fold right without zero. Ex: (foldr1 + '(3 2 6)) = (+ 3 (+ 2 6))
(define (foldr1 fun l)
  (if (null? (cdr l))
      (car l)
      (fun (car l) (foldr1 fun (cdr l)))))

; Fold left.
(define (foldl zero fun l)
  (foldr zero (lambda (x y) (fun y x)) (reverse l)))

; Fold left without zero.
(define (foldl1 fun l)
  (foldr1 (lambda (x y) (fun y x)) (reverse l)))


;;;;; Curryfication
; Passes only the first argument and returns a lambda that awaits the others.
(define (curry-1 f x)
  (lambda rest
    (apply f (cons x rest))))

(define (curry-2 f x y)
  (lambda rest
    (apply f (cons x (cons y rest)))))

(define (curry-args f l)
  (lambda rest
    (apply f (append l rest))))

(define (curry f)
  (lambda rest
    (apply f rest)))
    ; curry is same as (compose curry-1 apply)


;;;;; Composition
(define (compose-1 f g)
  (lambda args
     (f (apply g args))))

(define (compose f . glist )
  (lambda args
     (apply f (map (lambda (g) apply g args) glist))))

;;;;; Iteration
(define (iter-fun f n x)
  (if (zero? n) x (iter-fun f (n-1) (f x))))

(define (nth-iter f n) (curry-2 iter-fun f n))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Data structures ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;; Counters

(define (new-counter)
  (let ((n 0))
    (lambda ()
      (let ((val n))
      (set! n (1+ val))
      val))))

;;;;; List processing
; maps a function to a list of argument lists for that function
(define (map-args f l) (map (curry f) l))
(define (map-2 f l)
   (if (null? l) '()
       (cons (f (car l) (cadr l)) (map-2 f (cddr l)))))
(define (map-and pred lst)
  (or (null? lst)
      (and (pred (car lst))
           (map-and pred (cdr lst)))))
(define (map1 proc lst)
  (if (null? lst)
    '()
    (cons (proc (car lst)) (map1 proc (cdr lst)))))

;;;;; Filtering a list with a predicate -- keeps only good elements.
(define (filter pred lst)
  (if (null? lst)
    '()
    (let ((a (car lst))
	  (d (filter pred (cdr lst))))
      (if (pred a) (cons a d) d))))

(define (list-find p Y N l)
; finds the first element in list l matching predicate p
;    if found, runs function Y with arguments x,r,i,l
;    if not, runs function N with arguments i,l
; where
;    i is 0-based index of the element found (or list size if not found),
;    x is found element
;    r is rest of the list (starting from x)
;    l is the whole list
  (letrec ((find (lambda (r i)
    (cond
      ((null? l)   (N i l))
      ((p (car r)) (Y x r i l))
      (else        (find (cdr r) (+ i 1)))))))
    (find l 0)))


(define (list:last-pair l)
  (if (null? (cdr l)) l (list:last-pair (cdr l))))

(define (list:before-last-pair l)
  (if (null? (cddr l)) l (list:before-last-pair (cdr l))))

(define (list:copy l) (append l '()))


;;;;; flatten trees constructed with cons and list
(define (fringe x)
  (cond
    ((null? x) '())
    ((pair? x) (append (fringe (car x)) (fringe (cdr x))))
    (else (list x))))


;;;;; Simple records

(define (srec:new . rest)	  rest)
(define (srec:first s)		  (caar s))
(define (srec:second s)		  (cadar s))
(define (srec:third s)		  (caddar s))
(define (srec:set-first! s x)	  (set-car! (car s) x))
(define (srec:set-second! s x)	  (set-car! (cdar s) x))
(define (srec:set-third! s x)	  (set-car! (cddar s) x))


;;;;; Double queues
; implemented as a pair
; whose cdr points to the beginning of the stack element list, and
; whose car points to the last pair of the stack element list,
; or the original pair itself if the list is empty.
;
; * list are *not* copied in ->list and list->,
;  but should be if (un)queueing is not to clobber the list.
; * queueing in end of list is cheap, unqueueing is expensive
; * no checks are made - be careful !

(define (deq:list-> l)
   (let ((s (cons #f l)))
      (set-car! s (list:last-pair s))
      s))
(define (deq:new)		(deq:list-> '()))
(define (deq:->list x)		(cdr x))
(define (deq:empty? x)		(null? (cdr x)))
(define (deq:front x)		(cadr x))
(define (deq:push! x y)		(set-cdr! x (cons y (cdr x))))
(define (deq:pop! x)		(set-cdr! x (cddr x)))
(define (deq:rear x)		(caar x))
(define (deq:queue! x y)	(set-cdr! (car x) (cons y '()))
				(set-car! x (cdar x)))
(define (deq:unqueue! x)	(set-car! x (list:before-last-pair x))
				(set-cdr! (car x) '()))

;;;;; Traversal of structures

;;; A traversal function takes
; * a structure
; * a usual function to evaluate on every element,
;  one at a time, on every call of the function
; * an ending function to jump into when elements are exhausted
;  (possibly a continuation)
;;; injective traversals return each element once
;;; infinite structure traversals may never evaluate e
;;; solid traversals keep evaluating e when called beyond their end


;;; Using traversals
(define (traverse! t f)
   (while (t (lambda (s) (f s) #t) (lambda () #f))))

(define (traversal->istream s t c)
   (let*
      ((b? #f)
       (e? #f)
       (b  #f)
       (e (lambda () (set! b #f) (set! b? #f) (set! e? #t) #f))
       (u (lambda (x) (set! b x) (set! b? #t) (set! e? #f) #t))
       (n (t s u e))
       (fill (lambda () (or e? b? (n))))
       (get  (lambda () (fill) b))
       (next (lambda () (fill) (n)))
       (end? (lambda () e?)))
   (c get next end?)))


;;; Traversal of a list...
(define (list-traversal l)
   (lambda (f e)
      (if (null? l) (e)
	 (begin
	    (let ((x (car l)))
	       (set! l (cdr l))
	       (f x))))))

;(traverse! (list-traversal '(1 2 3 4 5)) print )


;;; Two-level Traversal of lists of structures...
(define (double-traversal t1 t2)
   (lambda (ss)
      (let*
	 ((t1ss  (t1 ss))	; traversing the main structure
	  (t2s   #f)		; traversing the current substructure
	  (step  #f))		; what to execute next time...
      (letrec
	 ((e2    (lambda (f e)  ; action at the end of current substructure
	            (lambda ()
		       (sstep f e))))
	  (f1    (lambda (f e)	; action during current substructure
		    (lambda (s)
		       (set! t2s (t2 s))
		       (set! step (lambda (f e) (t2s f (e2 f e))))
		       (step f e))))
	  (e1    (lambda (f e)  ; action at the end of the main structure
	            (lambda ()
	               (set! step (lambda (f e) e))
	               (e))))
          (sstep (lambda (f e)  
	            (t1ss (f1 f e) (e1 f e)))))
      (set! step sstep)
      (lambda (f e) (step f e))))))

(define list-list-traversal
   (double-traversal list-traversal list-traversal))

;(print "Hello, world !")
;
;(let ((L '((1 2) (3) () (4 5 6) ())))
;   (traverse! (list-traversal L) print)
;   (traverse! (list-list-traversal L) print))


;;;;; Looking-Up a structure
(define (look-up traversal predicate found not-found)
   (letrec
     ((loop (lambda (element)
         (if (predicate element) (found element) (next))))
      (next (lambda ()
                (traversal loop not-found))))
      (next)))

(define (->eqv? x)   (lambda (y) (eqv? x y)))	; (curry-1 eq? x)
(define (->eq? x)    (lambda (y) (eq? x y)))	; (curry-1 eqv? x)
(define (->equal? x) (lambda (y) (equal? x y)))	; (curry-1 equal? x)


;;;;; Association Tables

(define assoc:nil '())
(define (assoc:new) assoc:nil)
(define (assoc:cons key value table) (cons (cons key value) table))
(define (assoc:lookup table fun found not-found)
    (if (null? table) (not-found)
        (let* ((rest  (cdr table))
            (keva  (car table))
	    (key   (car keva)))
	(if (fun key)
	    (found keva)
	    (assoc:lookup rest fun found not-found)))))
(define (assoc:find table key found not-found)
	(assoc:lookup table (compare-to key) found not-found))
(define assoc:merge append)


;;;;;;;;;;;;;;;;;;; Arithmetical and Logical Operations ;;;;;;;;;;;;;;;;;;;;;

; We try to define versions that use a very small subset of
; Scheme's capabilities, so we can implement them all easily


;;;;; Integral powers
; standard definitions would be:
;	(define (intpow x n) (expt x n))
;	(define (*intpow a x n) (* a (intpow x n)))

(define (*intpow a x n)
  (letrec ((*pow (lambda (a x n)
		   (if (zero? n) a
		       (*pow (if (odd? n) (* a x) a)
			     (* x x)
			     (quotient n 2))))))
    (if (and (positive? n) (integer? n))
        (*pow a x n)
	(error "Only positive integer powers implemented."))))

(define (intpow x n) (*intpow 1 x n))



;;;;; Division
(define (ceil-/ n d) (ceiling (/ n d))) ; also (quotient (+ n d -1) d)


;;;;; Simulate Bitwise logical operations

(define (iquotient a b)
  (let ((q (quotient a b))
        (r (remainder a b)))
     (if (negative? r) (- q 1) q)))


; arithmetically-shift-right
(define (>> n i)
  (iquotient n (expt 2 i)))

; arithmetically-shift-left
(define (<< n i)
  (* n (expt 2 i)))


(define (all-bits-same? a)
  (or (zero? a) (eqv? a -1)))

(define (low-bit n)
  (if (even? n) 0 1))

(define (rest-bit n)
  (iquotient n 2))

(define (bitwise op)
  (lambda (a b)
    (if
      (and (all-bits-same? a) (all-bits-same? b))
      (op a b)
      (+ (op (low-bit a) (low-bit b))
	 (* 2 ((bitwise op) (rest-bit a) (rest-bit b)))))))

(define (assoc-op op neutral)
  (lambda args
    (foldr neutral op args)))


(define (or-1 a b)  (if (zero? a) b a))
(define (and-1 a b) (if (zero? a) a b))
(define (xor-1 a b) (if (zero? a) b (- a b)))

(define bitwise-or-2   (bitwise or-1))
(define bitwise-and-2  (bitwise and-1))
(define bitwise-xor-2  (bitwise xor-1))

(define bitwise-or  (assoc-op bitwise-or-2   0))
(define bitwise-and (assoc-op bitwise-and-2 -1))
(define bitwise-xor (assoc-op bitwise-xor-2  0))

(define (bitwise-not n) (- 0 1 n))

(define (trunc-bits n)
  (let ((mask (- (<< 1 n) 1)))
     (lambda (x) (bitwise-and-2 mask x))))


;;;;; Bit patterns
; these functions construct/destruct actual bit fields
; from/into actual integers.
; they do no symbolic stuff

(define (construct-fielded-value . field-list)
  (if (null? field-list) (throw-mismatch)
    (lambda fields
      (if (not (eqv? (length fields) (length field-list))) (throw-mismatch)
        (letrec
          ((cons-field (lambda (l f v)
	      (if (null? f) v
	        (let* ((n (car l))
	               (w (+ (<< v n) ((trunc-bits n) (car f)))))
	          (cons-field (cdr l) (cdr f) w))))))
           (cons-field field-list fields 0))))))

(define (destruct-fielded-value . field-list)
  (if (null? field-list) (throw-mismatch)
    (lambda (number)
       (letrec
          ((des-field (lambda (l r f)
	      (if (null? l) (if (zero? r) f (throw-mismatch))
	        (let* ((n (car l))
	               (b ((trunc-bits n) r))
		       (s (>> r n)))
	          (des-field (cdr l) s (cons b f)))))))
           (des-field (reverse field-list) number '())))))



;;;;;;;;;;;;;;;;;;;;;;;;;;; Symbol mangling ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (compare-to key) (lambda (k) (eqv? k key)))
(define (compare-car-to key) (lambda (k) (eqv? k key)))

(define (->string x)
   (cond
     ((string? x) x)
     ((symbol? x) (symbol->string x))
     ((number? x) (number->string x))
     ((char? x)   (make-string 1 x))
     (else "")))

(define (multi->string . args)
   (apply string-append (map ->string args)))

(define (make-symbol . args)
   (string->symbol (apply multi->string args)))


(define gensym-counter (new-counter))
(define (gensym . args)
   (apply make-symbol
	  (append '(G) args (gensym-counter))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Vectors ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (copy-vector v)
   (list->vector (vector->list v)))
